import { Stack } from "./stack.js";
import { InstructionPointer } from "./pointer.js";
import { EmptyCodeBoxError, FailedToParseCodeBoxError, StackOutOfBoundsError, WrongStackSizeError } from "./errors.js";
import { CodeBoxOptions, OrdinalDirection } from "./types.js";
import { charToDecimal, deepClone } from "./common.js";
import { Node } from "typescript";
import { InstructionDictionary } from "./instructions.js";

export class CodeBox {
    /**
     * The InstructionPointer for the CodeBox
     */
    public pointer: InstructionPointer = new InstructionPointer();
    /**
     * A list of stacks created in the CodeBox
     */
    public stacks: Array<Stack> = [];
    /**
     * The index of the currently selected stack
     */
    private currentStackIndex = 0;
    /**
     * The complete list of instructions for the CodeBox
     */
    private box: string[][] = [];
    /**
     * The furthest to the right the CodeBox extends
     */
    public maxBoxWidth: number = 0;
    /**
     * The furthest down the CodeBox extends
     */
    public maxBoxHeight: number = 0;
    /**
     * The current date/time
     *
     * Updated at the start of each command execution.
     */
    private dateTime: Date = new Date();
    /**
     * The original input for the code box
     *
     * Used when resetting the CodeBox.
     */
    private rawCodeBox: string[][] = [];
    /**
     * The original input for the starting stack
     * 
     * Used when resetting the CodeBox.
     */
    private initialStack: number[] = [];

    /**
     * The options for this CodeBox
     */
    public options: CodeBoxOptions = {
        debug: {
            print: {
                codeBox: false,
                stacks: false,
            }
        }
    }

    /**
     * The No OPeration instruction for Sunfish
     * @readonly
     */
    private NOP = " "; 
    /**
     * A list of movement and mirror related operators
     *
     * Needed for checking when the InstructionPointer's hasDove state is true
     */
    private MOVEMENT_AND_MIRRORS = [">", "<", "^", "v", "/", "\\", "|", "_", "#", "x", "`", "O"];

    /**
     * Create a Sunfish CodeBox
     * 
     * The CodeBox is the engine of the Sunfish interpreter.
     * 
     * @param rawCodeBox The initial CodeBox provided by the user
     * @param initialStack The initial stack of values provided by the user
     * @param options Options to modify the execution of the engine
     */
    constructor(rawCodeBox: string|string[] = [], initialStack: string = "", options: CodeBoxOptions|undefined = undefined) {
        // Catch for when being called from JS
        if (rawCodeBox.length === 0) {
            throw new EmptyCodeBoxError();
        }
        this.parseCodeBox(rawCodeBox);
        this.stacks.push(new Stack(initialStack));
        this.initialStack = this.stacks[0]!.values();
        if (options) {
            this.options = options;
        }

        // Set the default output callback to log to the console
        this.options.outputCallback ??= console.log;
        this.options.tickSleep ??= 0;
        if (this.options.tickSleep < 0) {
            throw new RangeError("the tickSleep option must be a float >= 0");
        }
        // Initialize the default debug options if none were set
        this.options.debug ??= {};
        this.options.debug.print ??= {}
        this.options.debug.print.logger ??= this.options.outputCallback;
        this.options.debug.print.codeBox ??= false;
        this.options.debug.print.stacks ??= false;
    }

    /**
     * Begin execution of the CodeBox
     */
    public run() {
        if (this.options.tickSleep) { return this.runTimed(); }
        let finished = false;

        while (!finished) {
            finished = this.swim();
        }
    }

    private runTimed() {
        let intervalID: NodeJS.Timeout;
        intervalID = setInterval(() => {
            if (this.swim()) {
                clearInterval(intervalID);
            }
        }, this.options.tickSleep as number * 1000);
    }

    /**
     * Execute one step inside the codebox
     *
     * @returns Returns true if the execution was halted
     */
    public swim(): boolean {
        this.logDebugs();

        const instruction = this.box[this.pointer.Y]![this.pointer.X] as string;
        this.dateTime = new Date(); 

        if (this.pointer.stringMode != 0 && charToDecimal(instruction) != this.pointer.stringMode) {
            this.stacks[this.currentStackIndex]!.push(charToDecimal(instruction));
        }
        else {
            const executionResult = this.execute(instruction);

            if (executionResult === true) {
                return executionResult
            }
            else if (executionResult !== false) {
                this.options.outputCallback!(executionResult);
            }
        }

        this.pointer.move();

        // Account for movement out of bounds
        if (this.pointer.X < 0) {
            this.pointer.X = this.maxBoxWidth - 1;
        }
        else if (this.pointer.X >= this.maxBoxWidth) {
            this.pointer.X = 0;
        }

        if (this.pointer.Y < 0) {
            this.pointer.Y = this.maxBoxHeight - 1;
        }
        else if (this.pointer.Y >= this.maxBoxHeight) {
            this.pointer.Y = 0;
        }

        return false;
    }

    /**
     * Process an instruction in the CodeBox
     *
     * @param instruction The instruction to process
     * @returns True if execution is halted, else false
     */
    private execute(instruction: string): boolean|number|string {
        let output: boolean|number|string = false;

        if (this.pointer.hasDove && this.MOVEMENT_AND_MIRRORS.indexOf(instruction) < 0) {
            return output;
        }

        switch (instruction) {
            case InstructionDictionary.NOP:
                break;
            // Numbers
            case InstructionDictionary.HEX_0:
            case InstructionDictionary.HEX_1:
            case InstructionDictionary.HEX_2:
            case InstructionDictionary.HEX_3:
            case InstructionDictionary.HEX_4:
            case InstructionDictionary.HEX_5:
            case InstructionDictionary.HEX_6:
            case InstructionDictionary.HEX_7:
            case InstructionDictionary.HEX_8:
            case InstructionDictionary.HEX_9:
            case InstructionDictionary.HEX_10:
            case InstructionDictionary.HEX_11:
            case InstructionDictionary.HEX_12:
            case InstructionDictionary.HEX_13:
            case InstructionDictionary.HEX_14:
            case InstructionDictionary.HEX_15:
                this.stacks[this.currentStackIndex]!.push(parseInt(instruction, 16));
                break;
            // Operators
            case InstructionDictionary.OPERATOR_ADD: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y + x);
                break;
            }
            case InstructionDictionary.OPERATOR_SUBTRACT: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y - x);
                break;
            }
            case InstructionDictionary.OPERATOR_MULTIPLY: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y * x);
                break;
            }
            case InstructionDictionary.OPERATOR_DIVIDE: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y / x);
                break;
            }
            case InstructionDictionary.OPERATOR_MODULO: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y % x);
                break;
            }
            case InstructionDictionary.OPERATOR_LESS_THAN: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y < x ? 1 : 0);
                break;
            }
            case InstructionDictionary.OPERATOR_GREATER_THAN: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y > x ? 1 : 0);
                break;
            }
            case InstructionDictionary.OPERATOR_EQUAL_TO: {
                const x = this.stacks[this.currentStackIndex]!.pop();
                const y = this.stacks[this.currentStackIndex]!.pop();
                this.stacks[this.currentStackIndex]!.push(y === x ? 1 : 0);
                break;
            }
            // String mode
            case InstructionDictionary.STRING_MODE_DOUBLE:
            case InstructionDictionary.STRING_MODE_SINGLE: {
                this.pointer.stringMode = !!this.pointer.stringMode ? 0 : charToDecimal(instruction);
                break;
            }
            // Dive, Rise, and Fisherman
            case InstructionDictionary.DIVE: {
                this.pointer.hasDove = true;
                break;
            }
            case InstructionDictionary.RISE: {
                this.pointer.hasDove = false;
                break;
            }
            case InstructionDictionary.FISHERMAN: {
                this.fisherman();
                break;
            }
            // Movement
            case InstructionDictionary.MOVE_UP: {
                this.moveUp();
                break;
            }
            case InstructionDictionary.MOVE_RIGHT: {
                this.moveRight();
                break;
            }
            case InstructionDictionary.MOVE_DOWN: {
                this.moveDown();
                break;
            }
            case InstructionDictionary.MOVE_LEFT: {
                this.moveLeft();
                break;
            }
            // Mirrors
            case InstructionDictionary.MIRROR_FORWARD: {
                this.reflectForward();
                break;
            }
            case InstructionDictionary.MIRROR_BACKWARD: {
                this.reflectBack();
                break;
            }
            case InstructionDictionary.MIRROR_HORIZONTAL: {
                this.horizontalMirror();
                break;
            }
            case InstructionDictionary.MIRROR_VERTICAL: {
                this.verticalMirror();
                break;
            }
            case InstructionDictionary.MIRROR_OMNI: {
                this.omniMirror();
                break;
            }
            case InstructionDictionary.MIRROR_RANDOM: {
                this.pointer.shuffleDirection();
                break;
            }
            // Trampolines
            case InstructionDictionary.TRAMPOLINE_CLASSIC: {
                this.pointer.move();
                break;
            }
            case InstructionDictionary.TRAMPLOINE_CONDITIONAL: {
                if (this.stacks[this.currentStackIndex]!.pop() === 0) { this.pointer.move(); }
                break;
            }
            // Stack manipulation
            case InstructionDictionary.STACK_REGISTER: {
                this.stacks[this.currentStackIndex]!.readWriteRegister();
                break;
            }
            case InstructionDictionary.STACK_DUPLICATE: {
                this.stacks[this.currentStackIndex]!.duplicate();
                break;
            }
            case InstructionDictionary.STACK_REMOVE: {
                this.stacks[this.currentStackIndex]!.remove();
                break;
            }
            case InstructionDictionary.STACK_SWAP_TWO: {
                this.stacks[this.currentStackIndex]!.swapTwo();
                break;
            }
            case InstructionDictionary.STACK_SWAP_THREE: {
                this.stacks[this.currentStackIndex]!.swapThree();
                break;
            }
            case InstructionDictionary.STACK_SHIFT_LEFT: {
                this.stacks[this.currentStackIndex]!.shiftLeft();
                break;
            }
            case InstructionDictionary.STACK_SHIFT_RIGHT: {
                this.stacks[this.currentStackIndex]!.shiftRight();
                break;
            }
            case InstructionDictionary.STACK_REVERSE: {
                this.stacks[this.currentStackIndex]!.reverse();
                break;
            }
            case InstructionDictionary.STACK_LENGTH: {
                this.stacks[this.currentStackIndex]!.pushLength();
                break;
            }
            case InstructionDictionary.STACK_SPLIT: {
                this.spliceStack(this.stacks[this.currentStackIndex]!.pop());
                break;
            }
            case InstructionDictionary.STACK_COLLAPSE: {
                this.collapseStack();
                break;
            }
            case InstructionDictionary.STACK_INDEX_INCREMENT: {
                this.currentStackIndex++;
                if (this.currentStackIndex > this.stacks.length) {
                    throw new StackOutOfBoundsError(this.currentStackIndex);
                }
                break;
            }
            case InstructionDictionary.STACK_INDEX_DECREMENT: {
                this.currentStackIndex--;
                if (this.currentStackIndex < 0) {
                    throw new StackOutOfBoundsError(this.currentStackIndex);
                }
                break;
            }
            // Output
            case InstructionDictionary.OUTPUT_DECIMAL: {
                output = this.stacks[this.currentStackIndex]!.pop();
                break;
            }
            case InstructionDictionary.OUTPUT_ASCII: {
                output = String.fromCharCode(this.stacks[this.currentStackIndex]!.pop());
                break;
            }
            // Time
            case InstructionDictionary.TIME_SLEEP: {
                setTimeout(this.run.bind(this), this.stacks[this.currentStackIndex]!.pop() * 100);
                this.pointer.move();
                output = true;
                break;
            }
            case InstructionDictionary.TIME_SECONDS: {
                this.stacks[this.currentStackIndex]!.push(this.dateTime.getUTCSeconds());
                break;
            }
            case InstructionDictionary.TIME_MINUTES: {
                this.stacks[this.currentStackIndex]!.push(this.dateTime.getUTCMinutes());
                break;
            }
            case InstructionDictionary.TIME_HOURS: {
                this.stacks[this.currentStackIndex]!.push(this.dateTime.getUTCHours());
                break;
            }
            // Code box manipulation
            case InstructionDictionary.CODE_BOX_FETCH: {
                this.pushFromCodeBox();
                break;
            }
            case InstructionDictionary.CODE_BOX_PLACE: {
                this.placeIntoCodeBox();
                break;
            }
            // Functions
            case InstructionDictionary.JUMP: {
                const [y, x] = this.stacks[this.currentStackIndex]!.popMany(2);
                this.pointer.X = x as number;
                this.pointer.Y = y as number;
                break;
            }
            case InstructionDictionary.JUMP_STORED: {
                // Store the pointer's coordinates on the stack beneath our current one
                const currCoords = new Stack([this.pointer.X, this.pointer.Y]);
                this.stacks.splice(this.currentStackIndex, 0, currCoords);
                this.currentStackIndex++;
                const [y, x] = this.stacks[this.currentStackIndex]!.popMany(2);
                this.pointer.X = x as number;
                this.pointer.Y = y as number;
                break;
            }
            case InstructionDictionary.RETURN: {
                // Remove the stored coordinates from the previous C call
                const storedCoords = this.stacks.splice(this.currentStackIndex - 1, 1).pop() as Stack;
                this.currentStackIndex--;
                this.pointer.Y = storedCoords.pop();
                this.pointer.X = storedCoords.pop();
                break;
            }
            case InstructionDictionary.EXIT:
                output = true;
                break;
            default:
                throw new Error(`Unknown instruction: ${instruction}`);
        }

        return output;
    }

    
    /**
     * Implement ^
     * 
     * Changes the swim direction upward
     */
    public moveUp() {
        this.pointer.changeDirection("NORTH");
    }
    /**
     * Implement >
     * 
     * Changes the swim direction rightward
     */
    public moveRight() {
        this.pointer.changeDirection("EAST");
    }
    /**
     * Implement v
     * 
     * Changes the swim direction downward
     */
    public moveDown() {
        this.pointer.changeDirection("SOUTH");
    }
    /**
     * Implement <
     * 
     * Changes the swim direction leftward
     */
    public moveLeft() {
        this.pointer.changeDirection("WEST");
    }

    /**
     * Implement `/`
     * 
     * Reflects the swim direction depending on its starting value
     */
    public reflectForward() {
        this.reflectAngle(true);
    }
    /**
     * Implement `\`
     * 
     * Reflects the swim direction depending on its starting value
     */
    public reflectBack() {
        this.reflectAngle(false);
    }

    /**
     * Reflect the direction of the InstructionPointer
     *
     * @param forward Whether the reflect was a forward slash or not
     */
    private reflectAngle(forward: boolean) {
        switch (this.pointer.getDirection()) {
            case "NORTH":
                forward ? this.moveRight() : this.moveLeft();
                break;
            case "EAST":
                forward ? this.moveUp() : this.moveDown();
                break;
            case "SOUTH":
                forward ? this.moveLeft() : this.moveRight();
                break;
            case "WEST":
                forward ? this.moveDown() : this.moveUp();
                break;
        }
    }

    /**
     * Implement `|`
     * 
     * Swaps the horizontal swim direction to its opposite
     */
    public horizontalMirror() {
        this.pointer.getDirection() == "EAST" ? this.moveLeft() : this.moveRight();
    }
    /**
     * Implement `_`
     * 
     * Swaps the horizontal swim direction to its opposite
     */
    public verticalMirror() {
        this.pointer.getDirection() == "NORTH" ? this.moveDown() : this.moveUp();
    }
    /**
     * Implement `#`
     * 
     * A combination of the vertical and the horizontal mirror
     */
    public omniMirror() {
        this.pointer.isMovingHorizontally() ? this.horizontalMirror() : this.verticalMirror();
    }

    /**
     * Implement `[`
     *
     * Takes X number of elements out of a stack and into a new stack
     *
     * This action creates a new stack, and places it on top of the one it was created from.
     * So, if you have three stacks, A, B, and C, and you splice a stack off of stack B,
     * the new order will be: A, B,  D, and C.
     *
     * @see {@link https://esolangs.org/wiki/Fish#Stacks ><> Documentation}
     *
     * @param spliceCount The number of elements to pop into a new stack
     */
    public spliceStack(spliceCount: number) {
        const stackCount = this.stacks[this.currentStackIndex]!.length();

        if (spliceCount > stackCount) {
            throw new WrongStackSizeError(spliceCount, stackCount);
        }

        // Need to reverse the popMany call to maintain order of elements
        const newStack = new Stack(this.stacks[this.currentStackIndex]!.popMany(spliceCount).reverse());

        // We're at the top of the stacks stack, so we can use .push
        if (this.currentStackIndex == this.stacks.length - 1) {
            this.stacks.push(newStack);
        }
        else {
            this.stacks.splice(this.currentStackIndex + 1, 0, newStack);
        }

        this.currentStackIndex++;
    }
    /**
     * Implement `]`
     * 
     * Collapses the current stack onto the one below it
     * If the current stack is the only one, it is replaced with a blank stack
     */
    public collapseStack() {
        // Undefined behavior collapsing the first stack down when there are other stacks available
        if (this.currentStackIndex == 0 && this.stacks.length != 1) {
            throw new Error();
        }

        if (this.currentStackIndex == 0) {
            this.stacks = [new Stack()];
        }
        else {
            const collapsed = this.stacks.splice(this.currentStackIndex, 1).pop() as Stack;
            this.currentStackIndex--;

            this.stacks[this.currentStackIndex]!.push(collapsed.values());
        }
    }

    /**
     * Implement `g`
     * 
     * Pops `y` and `x` from the stack, and then pushes the value of the character
     * at `[x, y]` in the code box.
     * 
     * NOP's and coords that are out of bounds are converted to 0.
     * 
     * Implements the behavior as defined by the original {@link https://gist.github.com/anonymous/6392418#file-fish-py-L306 ><>}, and not {@link https://github.com/redstarcoder/go-starfish/blob/master/starfish/starfish.go#L378 go-starfish}
     */
    public pushFromCodeBox() {
        const y = this.stacks[this.currentStackIndex]!.pop();
        const x = this.stacks[this.currentStackIndex]!.pop();

        let val: string = " ";
        try {
            val = this.box[y]![x] || " ";
        }
        catch (e) {
            val = " ";
        }

        const valParsed = val == " " ? 0 : charToDecimal(val);

        this.stacks[this.currentStackIndex]!.push(valParsed);
    }
    /**
     * Implement `p`
     * 
     * Pops `y`, `x`, and `v` off of the stack, and then places the string
     * representation of that value at `[x, y]` in the code box.
     */
    public placeIntoCodeBox() {
        const y = this.stacks[this.currentStackIndex]!.pop();
        const x = this.stacks[this.currentStackIndex]!.pop();
        const value = this.stacks[this.currentStackIndex]!.pop();

        while(y >= this.box.length) {
            this.box.push([]);
        }
        while(x >= this.box[y]!.length) {
            this.box[y]!.push(this.NOP);
        }

        this.equalizeBoxDimensions();

        this.box[y]![x] = String.fromCharCode(value);
    }

    /**
     * Implement `
     * 
     * Changes the swim direction based on the previous direction
     * @see https://esolangs.org/wiki/Starfish#Fisherman
     */
    public fisherman() {
        // Are we moving up or down?
        if (this.pointer.isMovingVertically()) {
            if (this.pointer.wasDirectionLeft) {
                this.moveLeft();
            }
            else {
                this.moveRight();
            }
        }
        else {
            if (this.pointer.onTheHook) {
                this.pointer.onTheHook = false;
                this.moveUp();
            }
            else {
                this.pointer.onTheHook = true;    
                this.moveDown();
            }
        }
    }

    /**
     * Log debug calls based on the the CodeBox's options
     */
    public logDebugs() {
        if (this.options.debug!.print!.codeBox) { this.options.debug!.print!.logger!(this.getCodeBoxAsString())}
        if (this.options.debug!.print!.stacks)  { this.options.debug!.print!.logger!(this.getStacksAsString())}
    }
    /**
     * Get the CodeBox in string form
     *
     * @returns A string containing the current state of the CodeBox
     */
    public getCodeBoxAsString(): string {
        let output = "";

        for (let y = 0; y < this.maxBoxHeight; y++) {
            for (let x = 0; x < this.maxBoxWidth; x++) {
                let instruction = this.box[y]![x];
                if (x == this.pointer.X && y == this.pointer.Y) {
                    instruction = `*${instruction}*`;
                }
                output += `${instruction} `;
            }
            output += "\n";
        }

        return output;
    }
    /**
     * Get the CodeBox's stacks as a string
     *
     * @returns A string representing a JSON object of the CodeBox's stacks' values
     */
    public getStacksAsString(): string {
        let output = "{\n";

        for (let i = 0; i < this.stacks.length; i++) {
            output += `\t${i}: ${JSON.stringify(this.stacks[i]!.values())},\n`;
        }

        output += "}";

        return output;
    }

    /**
     * Reset the CodeBox to how it was at the time of initializing
     */
    public reset() {
        this.box = deepClone(this.rawCodeBox);
        this.updateBoxDimensions();
        this.pointer = new InstructionPointer();
        this.stacks = [new Stack(this.initialStack)];
        this.dateTime = new Date();
    }

    /**
     * Parse an input into a useable CodeBox 
     */
    public parseCodeBox(toParse: string|string[]|string[][]) {
        let parsed = toParse;
        // If the original input is just a string, split it on all newlines
        if (typeof parsed == "string") {
            parsed = parsed.split(/\r?\n/);
        }
        // If the parsed input is now a 1-dimensional string array, split each row into individual chars
        if (Array.isArray(parsed) && typeof parsed[0] == "string") {
            parsed = (parsed as string[]).map((row: string) => row.split(""));
        }
        // Ensure that parsed is now a 2-dimensional string array
        if (!Array.isArray(parsed) || !Array.isArray(parsed[0])) {
            throw new FailedToParseCodeBoxError(toParse);
        }

        // Store the parsed box
        this.box = parsed as string[][];

        // Find the dimensions of the box
        this.updateBoxDimensions();

        // Fill in empty spaces
        this.equalizeBoxDimensions();

        // Store the original parsed CodeBox
        this.rawCodeBox = deepClone(this.box);
    }

    /**
     * Sets each row of the CodeBox to the same length
     * 
     * While not strictly necessary, as TypeScript/JavaScript can handle out-
     * of-bounds array calls, this does reduce some of the checks that would
     * otherwise be needed.
     */
    private equalizeBoxDimensions() {
        this.updateBoxDimensions();

        for (let row = 0; row < this.box.length; row++) {
            while (this.box[row]!.length != this.maxBoxWidth) {
                this.box[row]!.push(this.NOP);
            }
        }
    }

    /**
     * Updates the stored max height and width of the CodeBox
     */
    private updateBoxDimensions() {
        this.maxBoxHeight = this.box.length;

        let maxWidth = 0;
        for (const row of this.box) {
            if (row.length > maxWidth) { maxWidth = row.length; }
        }
        this.maxBoxWidth = maxWidth;
    }
};
