import type { Vector2, OrdinalDirection } from "./types.js"; 

export class InstructionPointer {
    /**
     * The current X coordinate of the InstructionPointer
     */
    public X = 0;
    /**
     * The current Y coordinate of the InstructionPointer
     */
    public Y = 0;
    /**
     * Possible vectors the InstructionPointer can move in
     */
    private directions: Record<OrdinalDirection, Vector2> = {
        NORTH: [ 0, -1],
        EAST:  [ 1,  0],
        SOUTH: [ 0,  1],
        WEST:  [-1,  0],
    };
    private currentDirection: OrdinalDirection = "EAST";
    /**
     * The current vector the Instruction Pointer is moving in
     * 
     * At the start of a CodeBox's execution, this will always be to the right
     */
    public currentVector = this.directions[this.currentDirection];
    /**
     * Was the InstructionPointer last moving to the left
     */
    public wasDirectionLeft = false;
    /**
     * Is the InstructionPointer currently under water
     * 
     * True if a `u` instruction has been reached. Returns to false once a
     * `O` instruction is reached.
     */
    public hasDove = false;
    /**
     * Is the InstructionPointer currently under the influence of the Fisherman
     * 
     * True if a ` instruction is reached. Returns to false if another one
     * is encountered.
     */
    public onTheHook = false;

    /**
     * Is the InstructionPointer currently in string mode
     * 
     * When either a `'` or a `"` is encountered during execution, the decimal
     * value of the character is stored here.
     * 
     * When this value is non-zero, any character encountered during execution
     * is treated as a value to be converted to ASCII and pushed onto the stack.
     * 
     * This value is reset to zero the next time a matching value is found.
     */
    public stringMode = 0;

    /**
     * Change the direction the InstructionPointer is moving in
     *
     * @param newDirection The new direction for the InstructionPointer to move
     */
    public changeDirection(newDirection: OrdinalDirection) {
        // Only update wasDirectionLeft when changing horizontal directions
        if (newDirection == "WEST" || newDirection == "EAST") {
            this.wasDirectionLeft = newDirection == "WEST";
        }
        this.currentDirection = newDirection;
        this.currentVector = this.directions[this.currentDirection];
    }

    /**
     * Get the label of the current direction the InstructionPointer is moving in
     */
    public getDirection(): OrdinalDirection {
        return this.currentDirection;
    }

    /**
     * Check if the InstructionPointer is moving horizontally
     */
    public isMovingHorizontally(): boolean {
        return this.currentVector[0] != 0;
    }
    /**
     * Check if the InstructionPointer is moving vertically
     */
    public isMovingVertically(): boolean {
        return !this.isMovingHorizontally();
    }

    /**
     * Implement `x`
     * 
     * Pseudo-randomly changes the direction of the InstructionPointer.
     */
    public shuffleDirection() {
        this.changeDirection(Object.keys(this.directions)[Math.floor(Math.random() * 4)] as OrdinalDirection);
    }

    /**
     * Move the InstructionPointer one step in its current direction
     */
    public move() {
        this.X += this.currentVector[0];
        this.Y += this.currentVector[1];
    }
};
