import { charToDecimal } from "./common.js";
import { EmptyStackError, WrongStackSizeError } from "./errors.js";

export class Stack {
    /**
     * The underlying array of numbers composing the stack
     */
    private stack: number[] = [];
    /**
     * The register for the stack
     */
    private register: number|null = null;

    /**
     * Create a Stack for a CodeBox to operate on
     *
     * @param initialValues A list of values to initialize the stack with
     */
    constructor(initialValues: string|number|number[] = []) {
        if (typeof initialValues == "string") {
            this.stack = this.parseStackString(initialValues);
        }
        else if (typeof initialValues == "number") {
            this.stack.push(initialValues);
        }
        else {
            this.stack = initialValues;
        }

        // Check to make sure the stack only has numbers
        this.stack = this.stack.filter((val) => !isNaN(val));
    }

    /**
     * Wrapper function for Array.prototype.push 
     */
    public push(newValue: number|number[]) {
        if(Array.isArray(newValue)) {
            this.stack.push(...newValue);
        }
        else {
            this.stack.push(newValue);
        }
    }

    /**
     * Wrapper function for Array.prototype.pop
     *
     * @returns {number}
     */
    public pop(): number {
        const value = this.stack.pop();
        if(value == undefined){ throw new EmptyStackError(); }

        return value;
    }

    /**
     * Pop N number of elements off the top of the stack
     *
     * @param count The amount of elements to pop off the stack
     * @returns An array containing the elements popped off
     */
    public popMany(count: number): number[] {
        const stackLength = this.stack.length;
        if (count > stackLength) {
            throw new WrongStackSizeError(count, stackLength);
        }

        return this.stack.splice(stackLength - count, count).reverse();
    }

    /**
     * Implement `}`
     * 
     * Shifts the entire stack leftward by one value
     */
    public shiftLeft() {
        const temp = this.stack.shift();

        if (temp !== undefined) {
            this.stack.push(temp);
        } else {
            throw new EmptyStackError();
        }
    }
    /**
     * Implement `{`
     * 
     * Shifts the entire stack rightward by one value
     */
    public shiftRight() {
        const temp = this.pop();
        
        this.stack.unshift(temp);
    }
    /**
     * Implement `$`
     * 
     * Swaps the top two values of the stack
     */
    public swapTwo() {
        if(this.stack.length < 2) { throw new WrongStackSizeError(2, this.stack.length); }
        if (this.stack.length == 2) { this.stack.reverse(); return; }

        const popped = this.stack.splice(this.stack.length - 2, 2);
        this.stack.push(...popped.reverse());
    }
    /**
     * Implement `@`
     * 
     * Swaps the top three values of the stack
     */
    public swapThree() {
        if(this.stack.length < 3) { throw new WrongStackSizeError(3, this.stack.length); }
        // Get the top three values
        const popped = this.stack.splice(this.stack.length - 3, 3);
        // Shift the elements to the right
        popped.unshift(popped.pop() as number);
        this.stack.push(...popped);
    }

    /**
     * Implement `:`
     * 
     * Duplicates the element on the top of the stack
     */
    public duplicate() {
        if (this.stack.length < 1) { throw new EmptyStackError(); }
        this.stack.push(this.stack[this.stack.length-1]!);
    }
    /**
     * Implements `~`
     * 
     * Removes the element on the top of the stack 
     */
    public remove() {
        if (this.stack.length < 1) { throw new EmptyStackError(); }
        this.stack.pop();
    }

    /**
     * Implement `r`
     * 
     * Reverses the entire stack
     */
    public reverse() {
        this.stack.reverse();
    }

    /**
     * Implement `l`
     * 
     * Pushes the length of the stack onto the top of the stack
     */
    public pushLength() {
        this.stack.push(this.stack.length);
    }

    /**
     * Implement `&`
     *
     * If the register currently has a value, that value is removed, and pushed
     * onto the stack.
     *
     * If the register is empty, a value is popped off the stack, and stored in
     * the register. 
     */
    public readWriteRegister() {
        if (this.register === null) {
            this.register = this.pop();
        }
        else {
            this.push(this.register);
            this.register = null;
        }
    }
    /**
     * Get the current value in the register
     */
    public getRegisterValue(): number|null {
        return this.register;
    }

    /**
     * Get all values in a stack
     */
    public values(): number[] {
        return this.stack;
    }

    /**
     * Get the length of the stack
     */
    public length(): number {
        return this.stack.length;
    }

    /**
     * Parses numbers out of a user-provided string to get a list of stack values
     *
     * @param toParse A string representing a list of stack values
     * @returns {number[]} 
     */
    private parseStackString(toParse: string): number[] {
        const parsingSeparator = new RegExp(`(["'].+?["']|\d+)|\\s|,`, "g");
        const parsedValues: number[] = [];

        const splitValues: Array<string> = toParse.split(parsingSeparator).map((value) => value?.trim()).filter((value) => value?.length);

        for (const val of splitValues) {
            if (!isNaN(parseInt(val, 10))) {
                parsedValues.push(parseInt(val, 10));
            }
            else {
                let chars = val.substring(1, val.length - 1).split('');
                parsedValues.push(...chars.map((c) => charToDecimal(c)));
            }
        }

        return parsedValues;
    }
};
